#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.core.tools
      @file: text_tools.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from hspylib.core.tools.dict_tools import get_or_default
from typing import Any, List, Optional, Tuple

import os
import random
import re
import string


def elide_text(text: str, width: int, elide_with: str = "...") -> str:
    """Return a copy of the string eliding the text case the string is bigger than the specified length."""
    return text if len(text) <= width else text[: width - len(elide_with)] + elide_with


def cut(text: str, field: int, separator: str = " ") -> Tuple[Optional[str], tuple]:
    """Return a new string cut out from the given text."""
    result = tuple(re.split(rf"{separator}+", text))
    return get_or_default(result, field), result[:field]


def random_string(choices: List[str], length: int, weights: List[int] = None) -> str:
    """Return a new random string matching choices and length."""
    return "".join(random.choices(choices, weights or [1] * len(choices), k=length))


def justified_left(text: str, width: int, fill: str = " ") -> str:
    """Return a copy of the string justified left."""
    return text.ljust(width, fill)


def justified_center(text: str, width: int, fill: str = " ") -> str:
    """Return a copy of the string justified center."""
    return text.center(width, fill)


def justified_right(text: str, width: int, fill: str = " ") -> str:
    """Return a copy of the string justified right."""
    return text.rjust(width, fill)


def uppercase(text: str) -> str:
    """Return a copy of the string converted to upper case."""
    return text.upper()


def lowercase(text: str) -> str:
    """Return a copy of the string converted to lower case."""
    return text.lower()


def camelcase(text: str, separator: str = " |-|_", upper: bool = False) -> str:
    """Return a copy of the string converted to camel case.
    Ref:https://en.wikipedia.org/wiki/Letter_case :: Camel case"""
    s = re.sub(rf"({separator})+", " ", text)
    s = s.title()
    s = re.sub(rf"({separator})+", "", s)
    fnc = getattr(s[0], "lower" if not upper else "upper")
    return "".join([fnc(), s[1:]])


def snakecase(text: str, separator: str = "-", screaming: bool = False) -> str:
    """Return a copy of the string converted to snake case.
    Ref:https://en.wikipedia.org/wiki/Letter_case :: Snake case
    """
    s = re.sub("([A-Z][a-z]+)", r" \1", re.sub("([A-Z]+)", r" \1", text.replace(separator, " ")))
    text = "_".join(s.split())
    fnc = getattr(text, "lower" if not screaming else "upper")
    return fnc()


def kebabcase(text: str, separator: str = " |-|_", train: bool = False) -> str:
    """Return a copy of the string converted to kebab case.
    Ref:https://en.wikipedia.org/wiki/Letter_case :: Kebab case
    """
    s = re.sub("([A-Z][a-z]+)", r" \1", re.sub("([A-Z]+)", r" \1", text.replace(separator, " ")))
    text = "-".join(s.split())
    fnc = getattr(text, "lower" if not train else "upper")
    return fnc()


def titlecase(text: str, separator: str = " |-|_", skip_length: int = 0) -> str:
    """Return a copy of the string converted to title case.
    Ref:https://en.wikipedia.org/wiki/Letter_case :: Title case
    """
    s = re.sub(rf"({separator})+", " ", text)
    s = " ".join([string.capwords(word) if len(word) > skip_length else word.lower() for word in s.split(" ")])
    return s


def environ_name(property_name: str) -> str:
    """Retrieve the environment name of the specified property name.
    :param property_name: the name of the property using space, dot or dash notations
    """
    return re.sub("[ -.]", "_", property_name).upper()


def strip_escapes(text: str) -> str:
    """Return a copy of the string stripping out all ansi escape 'ESC[' codes from it.
    Ref:https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
    """
    ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    return ansi_escape.sub("", text)


def strip_linebreaks(text: str, re_exp: str = r"(\n|\r|\n\r)+") -> str:
    """Return a copy of the string stripping out all line breaks from it."""
    return re.sub(re.compile(rf"{re_exp}"), "", text)


def strip_extra_spaces(text: str, re_exp: str = r"\s+", trim: bool = False) -> str:
    """Return a copy of the string stripping out all components spaces 2+ from it."""
    s = re.sub(re.compile(rf"{re_exp}"), " ", text)
    return s if not trim else s.strip()


def split_and_filter(text: str, regex_filter: str = ".*", delimiter: str = os.linesep) -> List[str]:
    """Split the string using the delimiter and filter using the specified regex filter
    :param text: The string to be split
    :param regex_filter: The regex to filter the string
    :param delimiter: The delimiter according which to split the string
    :return:
    """
    return list(filter(re.compile(regex_filter).search, text.split(delimiter)))


def json_stringify(json_text: str) -> str:
    """Return a copy of the json text stripping any line breaks or formatting from it and also quoting any existing
    double quotes."""
    return strip_extra_spaces(strip_linebreaks(json_text)).replace('"', '\\"')


def eol(current_index: int, split_len: int, line_sep: str = os.linesep, word_sep: str = " ") -> str:
    """Give the line separator character (hit) or an empty string (miss) according to the splitting length and
    current index."""
    return line_sep if (current_index + 1) % split_len == 0 else word_sep


def ensure_endswith(text: str, end_str: str) -> str:
    """Ensure the string ends with the given end string"""
    return text if text.endswith(end_str) else text + end_str


def ensure_startswith(text: str, start_str: str) -> str:
    """Ensure the string starts with the given start string"""
    return text if text.startswith(start_str) else start_str + text


def quote(value: Any) -> str:
    """Quote or double quote the value according to the value type."""
    return (
        str(value)
        if not isinstance(value, str)
        else f'"{value}"'
        if value.startswith("'") and value.endswith("'")
        else f"'{value}'"
    )


def last_index_of(text: str, substring: str) -> int:
    """Return the last index of substring or -1 if substring was not found in text."""
    try:
        return text.rindex(substring)
    except ValueError:
        return -1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.cli.vt100
      @file: terminal.py
   @created: Tue, 11 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from abc import ABC
from hspylib.core.enums.charset import Charset
from hspylib.core.tools.commons import sysout
from hspylib.modules.application.exit_status import ExitStatus
from hspylib.modules.cli.keyboard import Keyboard
from typing import Optional, Tuple

import logging as log
import os
import platform
import select
import shlex
import signal
import subprocess


class Terminal(ABC):
    """Utility class to provide execution of commands on a terminal."""

    @staticmethod
    def shell_exec(cmd_line: str, **kwargs) -> Tuple[Optional[str], ExitStatus]:
        """Execute command with arguments and return it's run status."""
        try:
            if "stdout" in kwargs:
                del kwargs["stdout"]  # Deleted since we use our own stream
            if "stderr" in kwargs:
                del kwargs["stderr"]  # Deleted since we use our own stream
            log.info("Executing shell command: %s", cmd_line)
            cmd_args = list(filter(None, shlex.split(cmd_line)))
            output = subprocess.check_output(cmd_args, **kwargs).decode(Charset.UTF_8.val)
            log.info("Execution result: %s", ExitStatus.SUCCESS)
            return output.strip() if output else None, ExitStatus.SUCCESS
        except subprocess.CalledProcessError as err:
            log.error("Command failed: %s => %s", cmd_line, err)
            return None, ExitStatus.FAILED

    @staticmethod
    def shell_poll(cmd_line: str, **kwargs) -> None:
        """Execute command with arguments and continuously poll it's output."""
        if "stdout" in kwargs:
            del kwargs["stdout"]  # Deleted since we use our own stream
        if "stderr" in kwargs:
            del kwargs["stderr"]  # Deleted since we use our own stream
        try:
            log.info("Polling shell command: %s", cmd_line)
            cmd_args = list(filter(None, shlex.split(cmd_line)))
            with subprocess.Popen(
                cmd_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, preexec_fn=os.setsid, **kwargs
            ) as proc:
                process = select.poll()
                process.register(proc.stdout)
                while not Keyboard.kbhit():
                    if poll_obj := process.poll(0.3):
                        line = bytes(proc.stdout.readline()).decode(Charset.UTF_8.val).strip()
                        sysout(line)
                        log.debug("Polling returned: %s", str(poll_obj))
                os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
        except (InterruptedError, KeyboardInterrupt):
            log.warning("Polling process has been interrupted command='%s'", cmd_line)
        except subprocess.CalledProcessError as err:
            log.error("Command failed: %s => %s", cmd_line, err)

    @staticmethod
    def open_file(filename: str) -> None:
        """Open the specified file using the default editor."""
        my_os = os.environ.get("HHS_MY_OS", platform.system())
        if "Darwin" == my_os:
            Terminal.shell_exec(f"open {filename}")
        elif "Linux" == my_os:
            Terminal.shell_exec(f"xdg-open {filename}")
        else:
            raise NotImplementedError(f"OS '{my_os}' is not supported")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.cli.vt100
      @file: vt_utils.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from hspylib.core.exception.exceptions import NotATerminalError
from hspylib.core.tools.commons import is_debugging, sysout
from hspylib.modules.cli.vt100.vt_100 import Vt100
from typing import Any, Tuple

import logging as log
import os
import re
import sys
import termios
import tty


def screen_size() -> Tuple[int, ...]:
    """Retrieve the size of the terminal"""
    if not sys.stdout.isatty():
        log.error(NotATerminalError("screen_size:: Requires a terminal (TTY)"))
        return 0, 0

    return tuple(map(int, os.popen("stty size").read().split()))


# Solution taken from:
# https://stackoverflow.com/questions/46651602/determine-the-terminal-cursor-position-with-an-ansi-sequence-in-python-3
def get_cursor_position() -> Tuple[int, int]:
    """Get the terminal cursor position"""
    pos = 0, 0

    if is_debugging() or not sys.stdout.isatty():
        log.error(NotATerminalError("get_cursor_position:: Requires a terminal (TTY)"))
        return pos

    buf = ""
    stdin = sys.stdin.fileno()
    attrs = termios.tcgetattr(stdin)  # Save terminal attributes

    try:
        tty.setcbreak(stdin, termios.TCSANOW)
        sys.stdout.write(Vt100.get_cursor_pos())  # Query the terminal
        sys.stdout.flush()
        while True:
            buf += sys.stdin.read(1)
            if buf[-1] == "R":
                break
        # If the response is 'Esc[r;cR'
        if matches := re.match(r"^\x1b\[(\d*);(\d*)R", buf):
            groups = matches.groups()
            pos = int(groups[0]), int(groups[1])
    finally:
        # Reset terminal attributes
        termios.tcsetattr(stdin, termios.TCSANOW, attrs)

    return pos


def set_enable_echo(enabled: bool = True) -> None:
    """Enable echo in the terminal
    :param enabled: whether is enabled or not
    """
    if not sys.stdout.isatty():
        log.error(NotATerminalError("set_enable_echo:: Requires a terminal (TTY)"))
        return

    os.popen(f"stty {'echo -raw' if enabled else 'raw -echo min 0'}").read()


def set_auto_wrap(auto_wrap: bool = True) -> None:
    """Set auto-wrap mode in the terminal
    :param auto_wrap: whether auto_wrap is set or not
    """
    if not sys.stdout.isatty():
        log.warning(NotATerminalError("set_enable_echo:: Requires a terminal (TTY)"))
        return

    sysout(Vt100.set_auto_wrap(auto_wrap), end="")


def set_show_cursor(show_cursor: bool = True) -> None:
    """Show or hide cursor in the terminal
    :param show_cursor: whether to show or hide he cursor
    """
    if not sys.stdout.isatty():
        log.warning(NotATerminalError("set_enable_echo:: Requires a terminal (TTY)"))
        return

    sysout(Vt100.set_show_cursor(show_cursor), end="")


def erase_line(mode: int = 2) -> None:
    """Erase current line"""
    sysout(f"%EL{mode}%\r", end="")


def save_cursor() -> None:
    """Save cursor position and attributes"""
    sysout(Vt100.save_cursor(), end="")


def clear_screen() -> None:
    """Clear terminal"""
    sysout("%ED2%", end="")
    sysout("%HOM%", end="")


def restore_cursor() -> None:
    """Restore cursor position and attributes"""
    sysout(Vt100.restore_cursor(), end="")


def alternate_screen(enable: bool = True) -> None:
    """Switch to the alternate screen buffer on/off"""
    sysout(f"%SC{'A' if enable else 'M'}%", end="")


def restore_terminal() -> None:
    """Clear the terminal and restore default attributes"""
    set_auto_wrap()
    set_show_cursor()
    set_enable_echo()


def exit_app(exit_code: int = None, frame: Any = None, exit_msg: str = "") -> None:
    """Exit the application. Commonly called by hooked signals."""
    restore_terminal()
    sysout(f"%MOD(0)%" f"{exit_msg if exit_msg else ''}")
    sys.exit(exit_code)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.application.argparse
      @file: arguments_builder.py
   @created: Thu, 14 Feb 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from argparse import ArgumentParser
from functools import partial
from hspylib.modules.application.argparse.parser_action import ParserAction
from typing import Any, Union


class ArgumentsBuilder:
    """Class to provide an argument parser builder."""

    def __init__(self, arg_parser: ArgumentParser):
        self._arg_parser = arg_parser

    def argument(
        self,
        name: str,
        help_string: str = None,
        choices: list = None,
        nargs: Union[str, int] = None,
        action: ParserAction = ParserAction.STORE) -> "ArgumentsBuilder":
        """Assign a new argument to the parser."""

        add_argument = partial(
            self._arg_parser.add_argument,
            dest=name,
            help=help_string or f"the {name}",
            action=action.value)

        if action == ParserAction.STORE:
            add_argument(choices=choices, nargs=nargs)
        else:
            add_argument()

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.core.tools
      @file: git_tools.py
   @created: Nov 14, 2019
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from hspylib.modules.application.exit_status import ExitStatus
from hspylib.modules.cli.terminal import Terminal
from typing import Tuple


class GitTools:
    """TODO"""

    @staticmethod
    def top_level_dir() -> Tuple[str, ExitStatus]:
        """TODO"""
        return Terminal.shell_exec("git rev-parse --show-toplevel")

    @staticmethod
    def current_branch() -> Tuple[str, ExitStatus]:
        """TODO"""
        return Terminal.shell_exec("git symbolic-ref --short HEAD")

    @staticmethod
    def changelog(from_tag: str, to_tag: str) -> Tuple[str, ExitStatus]:
        """TODO"""
        return Terminal.shell_exec(f"git log --oneline --pretty='format:%h %ad %s' --date=short {from_tag}^..{to_tag}^")

    @staticmethod
    def unreleased() -> Tuple[str, ExitStatus]:
        """TODO"""
        latest_tag = Terminal.shell_exec("git describe --tags --abbrev=0 HEAD^")
        return Terminal.shell_exec(f"git log --oneline --pretty='format:%h %ad %s' --date=short '{latest_tag}'..HEAD")

    @staticmethod
    def release_date(tag_name: str) -> Tuple[str, ExitStatus]:
        """TODO"""
        return Terminal.shell_exec(f"git log -1 --pretty='format:%ad' --date=short {tag_name}")

    @staticmethod
    def tag_list() -> Tuple[str, ExitStatus]:
        """TODO"""
        return Terminal.shell_exec("git log --tags --simplify-by-decoration --pretty='format:%ci %d'")

    @staticmethod
    def create_tag(version: str, commit_id: str = "HEAD", description: str = None) -> Tuple[str, ExitStatus]:
        """TODO"""
        return Terminal.shell_exec(f"git tag -a v{version} {commit_id} -m '{description or version}'")

    @staticmethod
    def search_logs(filter_by: str = ".*") -> Tuple[str, ExitStatus]:
        """TODO"""
        return Terminal.shell_exec(f"git log --grep='{filter_by}' --pretty=format:'%h %ad %s' --date=short")

    @staticmethod
    def show_file(filename: str, commit_id: str = "HEAD") -> Tuple[str, ExitStatus]:
        return Terminal.shell_exec(f"git show {commit_id}:{filename}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.core.metaclass
      @file: singleton.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""

from abc import ABCMeta
from hspylib.core.exception.exceptions import HSBaseException
from hspylib.core.preconditions import check_not_none
from typing import Any, Type, TypeVar, Union

import logging as log
import sys

SINGLETON = TypeVar("SINGLETON", bound=Union[Type, "Singleton"])


class Singleton(Type):
    """Singleton pattern is a software design pattern that restricts the instantiation of a class to a singular
    instance. This metaclass enables a class to be singleton."""

    _instances = {}

    # pylint: disable=bad-mcs-method-argument
    def __call__(mcs, *args, **kwargs) -> Any:
        """Invoke the class constructor or return the instance if it exists."""
        if not Singleton.has_instance(mcs):
            try:
                instance = super().__call__(*args, **kwargs)
                check_not_none(instance, f"Unable to create Singleton instance: {mcs}")
                setattr(mcs, "INSTANCE", instance)
                Singleton._instances[mcs.__name__] = instance
                log.debug("Created a new Singleton instance: %s.%s", mcs.__module__, mcs.__name__)
                return instance
            except Exception as err:
                raise HSBaseException(f"Failed to create singleton instance: '{mcs.__name__}'", err) from err
        return Singleton._instances[mcs.__name__]

    @classmethod
    def has_instance(cls, clazz: SINGLETON) -> bool:
        """Whether the class has an instance or not."""
        return clazz.__name__ in cls._instances

    @classmethod
    def del_instance(cls, clazz: SINGLETON) -> None:
        """Deletes the singleton instance. This method should be used only for testing purposes."""
        if any(m in sys.modules for m in ["unittest", "pytest"]):
            if Singleton.has_instance(clazz):
                log.warning("Deleted an existing Singleton instance: %s.%s", cls.__module__, cls.__name__)
                del cls._instances[clazz.__name__]
                delattr(clazz, "INSTANCE")
                del clazz
            else:
                raise HSBaseException(f"Failed to delete singleton instance: '{clazz.__name__}' was not found")
        else:
            raise HSBaseException("This method is only available for testing purposes (cleanup).")


class AbstractSingleton(Singleton, ABCMeta):
    """This metaclass mixes Singleton and Abstract class."""

    def __call__(cls, *args, **kwargs) -> Any:
        return super().__call__(*args, **kwargs)



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.187
#
# Package: main.hspylib.core.enums
"""Package initialization."""

__all__ = [
  'charset', 
  'content_type', 
  'enumeration', 
  'http_code', 
  'http_method'
]
__version__ = '0.11.187'



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.187
#
# Package: main.hspylib.modules
"""Package initialization."""

__all__ = [
  'application', 
  'cache', 
  'cli', 
  'eventbus', 
  'fetch', 
  'security'
]
__version__ = '0.11.187'



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.187
#
# Package: main.hspylib.core.config
"""Package initialization."""

__all__ = [
  'app_config', 
  'parser_factory', 
  'properties'
]
__version__ = '0.11.187'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.application
      @file: application.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
import argparse
import atexit
import logging as log
import os
import sys
import traceback
from abc import abstractmethod
from textwrap import dedent
from typing import Any, Optional

from hspylib.core.config.app_config import AppConfigs
from hspylib.core.exception.exceptions import ApplicationError
from hspylib.core.metaclass.singleton import AbstractSingleton
from hspylib.core.preconditions import check_state
from hspylib.core.tools.commons import hook_exit_signals, log_init, syserr, sysout
from hspylib.core.tools.text_tools import camelcase
from hspylib.core.zoned_datetime import now
from hspylib.modules.application.argparse.argument_parser import HSArgumentParser
from hspylib.modules.application.argparse.arguments_builder import ArgumentsBuilder
from hspylib.modules.application.argparse.chained_arguments_builder import ChainedArgumentsBuilder
from hspylib.modules.application.argparse.options_builder import OptionsBuilder
from hspylib.modules.application.exit_hooks import ExitHooks
from hspylib.modules.application.exit_status import ExitStatus
from hspylib.modules.application.version import Version


class Application(metaclass=AbstractSingleton):
    """HsPyLib application framework. This is the base class for the HsPyLib applications."""

    INSTANCE = None

    @staticmethod
    def _help_formatter(prog) -> argparse.HelpFormatter:
        return argparse.RawDescriptionHelpFormatter(
            prog,
            indent_increment=2,
            max_help_position=27
        )

    @staticmethod
    def exit(signum: int = 0, frame: Any = None, clear_screen: bool = False) -> None:
        """
        Handle interruptions to shutdown gracefully
        :param signum: The signal number or the exit http_code
        :param frame: The frame raised by the signal
        :param clear_screen: Whether to clean the screen before execution or not
        """
        if frame:
            log.warning("Signal handler hooked signum=%d frame=%s", signum, frame)
            exit_status = ExitStatus.ABORTED
        else:
            exit_status = ExitStatus.of(signum)
        if clear_screen:
            sysout("%ED2%%HOM%")

        sys.exit(exit_status.value)

    def __init__(
        self,
        name: str,
        version: Version,
        description: str = None,
        usage: str = None,
        epilog: str = None,
        resource_dir: str = None,
        log_dir: str = None,
    ):
        log.captureWarnings(True)
        hook_exit_signals(Application.exit)

        self.exit_hooks = ExitHooks(self._cleanup)
        self.exit_hooks.hook()
        self._run_dir = os.getcwd()
        self._args = {}
        self._exit_code = ExitStatus.NOT_SET
        self._app_name = name
        self._app_version = version
        self._app_description = description
        self._arg_parser = HSArgumentParser(
            exit_on_error=False,
            prog=name,
            allow_abbrev=False,
            formatter_class=self._help_formatter,
            description=dedent(description or ""),
            usage=usage,
            epilog=dedent(epilog or ""),
        )
        self._arg_parser.add_argument("-v", "--version", action="version", version=f"%(prog)s v{self._app_version}")

        # Initialize application configs
        if os.path.exists(f"{resource_dir}"):
            self.configs = AppConfigs(resource_dir=resource_dir)
        elif not resource_dir and os.path.exists(f"{self._run_dir}/resources/application.properties"):
            self.configs = AppConfigs(resource_dir=f"{self._run_dir}/resources")
        else:
            log.debug('Resource dir "%s" not found. AppConfigs will not be available!', resource_dir or "<none>")

        # Initialize application logs
        self._log_file = f"{log_dir or os.getenv('HHS_LOG_DIR', os.getcwd())}/{camelcase(name)}.log"
        check_state(log_init(self._log_file), "Unable to initialize logging. log_file={}", self._log_file)

    def run(self, *params, **kwargs) -> None:
        """Main entry point handler."""
        today = now()
        no_exit = "no_exit" in kwargs
        log.info("Application %s started %s", self._app_name, today)
        try:
            atexit.register(self._cleanup)
            self._setup_arguments()
            self._args = self._arg_parser.parse_args(*params)
            log.debug("Command line arguments: %s", str(self._args))
            self._exit_code = self._main(*params, **kwargs)
        except argparse.ArgumentError as err:
            log.error("Application failed to execute %s => %s", today, err)
            self.usage(ExitStatus.FAILED, no_exit=True)
            syserr(f"\n### Error {self._app_name} -> {err}\n\n")
            raise ApplicationError(f"Application failed to execute => {err}") from err
        except Exception as err:
            _, _, tb = sys.exc_info()
            if tb:
                traceback.print_exc(file=sys.stderr)
            log.error("Application execution failed %s => %s", today, err)
            self._exit_code = ExitStatus.ERROR
            raise ApplicationError(f"Application execution failed => {err}") from err
        finally:
            log.info("Application %s finished %s", self._app_name, today)
            if self._exit_code == ExitStatus.NOT_SET:
                _, code, _ = sys.exc_info()
                self._exit_code = ExitStatus.of(code)
            if not no_exit:
                Application.exit(self._exit_code.val)

    def usage(self, exit_code: ExitStatus = ExitStatus.SUCCESS, no_exit: bool = False) -> None:
        """Display the usage message and exit with the specified http_code ( or zero as default )
        :param no_exit: Do no exit the application on usage call
        :param exit_code: The exit http_code
        """
        self._exit_code = exit_code
        self._arg_parser.print_help(sys.stderr if self._exit_code != ExitStatus.SUCCESS else sys.stdout)
        if not no_exit:
            Application.exit(self._exit_code.val)

    def version(self) -> str:
        """Return the application version."""
        return str(self._app_version)

    def name(self) -> str:
        """Return the application name."""
        return self._app_name

    def get_arg(self, name: str) -> Optional[Any]:
        """Get the argument value specified by name."""
        return getattr(self._args, name) \
            if self._args and hasattr(self._args, name) \
            else None

    def _with_options(self) -> "OptionsBuilder":
        """Add options to the application."""
        return OptionsBuilder(self._arg_parser)

    def _with_arguments(self) -> "ArgumentsBuilder":
        """Add arguments to the application."""
        return ArgumentsBuilder(self._arg_parser)

    def _with_chained_args(self, arg: str, arg_help: str = None) -> "ChainedArgumentsBuilder":
        """Add chained arguments to the application."""
        return ChainedArgumentsBuilder(self._arg_parser, arg, arg_help)

    @abstractmethod
    def _setup_arguments(self) -> None:
        """Initialize application parameters and options."""

    @abstractmethod
    def _main(self, *params, **kwargs) -> ExitStatus:
        """Execute the application's main statements."""

    @abstractmethod
    def _cleanup(self) -> None:
        """Execute http_code cleanup before exiting."""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.cache
      @file: ttl_cache.py
   @created: Wed, 7 Dec 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from hspylib.core.enums.charset import Charset
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.preconditions import check_not_none
from hspylib.core.tools.commons import safe_delete_file
from hspylib.modules.cache.ttl_keyring_be import TTLKeyringBE
from hspylib.modules.security.security import b64_decode, b64_encode
from threading import Lock
from typing import Generic, Optional, TypeVar

import ast
import keyring
import os
import tempfile

T = TypeVar("T")


class TTLCache(Generic[T], metaclass=Singleton):
    """Class to provide a cache with time-to-live timeout."""

    CACHE_SERVICE = "HS-CACHE-SERVICE"

    def __init__(self, ttl_minutes: int = 15, ttl_seconds: int = 0) -> None:
        super().__init__()
        self._lock = Lock()
        keyring.set_keyring(TTLKeyringBE(ttl_minutes, ttl_seconds, safe_delete_file))

    def save(self, key: str, entry: T) -> str:
        """Save an entry identified by key containing the given value."""
        check_not_none(key, entry)
        with self._lock:
            with tempfile.NamedTemporaryFile(delete=False, mode="w", encoding=Charset.UTF_8.val) as f_temp:
                content = b64_encode(f"{repr(entry)}")
                f_temp.write(content)
                keyring.set_password(self.CACHE_SERVICE, key, f_temp.name)
                return f_temp.name

    def read(self, key: str) -> Optional[T]:
        """Read an entry identified by key."""
        check_not_none(key)
        with self._lock:
            cache_name = keyring.get_password(self.CACHE_SERVICE, key)
            if not cache_name or not os.path.exists(cache_name):
                return None
            try:
                with open(cache_name, encoding=Charset.UTF_8.val) as f_cache:
                    content = b64_decode(f_cache.readline())
                    return ast.literal_eval(content)
            except (ValueError, TypeError, SyntaxError):
                pass  # Just ignore those exceptions.
        return None

    def delete(self, key: str) -> None:
        """Delete an entry identified by key."""
        check_not_none(key)
        with self._lock:
            keyring.delete_password(self.CACHE_SERVICE, key)
